import socket
import logging

from OpenSSL import SSL
import certifi


log = logging.getLogger(__name__)


def get_host_certificate(host, port=443):
  ip_addr = socket.gethostbyname(host)
  sock = socket.socket()
  context = SSL.Context(SSL.TLSv1_METHOD)
  context.set_options(SSL.OP_NO_SSLv2)
  context.load_verify_locations(certifi.where(), None)
  ssl_sock = SSL.Connection(context, sock)
  ssl_sock.connect((ip_addr, port))
  ssl_sock.do_handshake()
  return ssl_sock.get_peer_certificate()



#!/usr/bin/env python

"""List contents of a certificate."""

import argparse
import logging
import ssl
import sys

from OpenSSL import crypto
import dateutil.parser
import tabulate

from py509 import client


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def stringify_version(v):
  try:
    return {
      ssl.PROTOCOL_SSLv2: 'SSLv2',
      ssl.PROTOCOL_SSLv23: 'SSLv23',
      ssl.PROTOCOL_SSLv3: 'SSLv3',
      ssl.PROTOCOL_TLSv1: 'TLSv1',
    }[v]
  except KeyError:
    return '???'


def stringify_subject(s):
  return 'CN=%(CN)s' % dict(s)


def main():

  #parser = argparse.ArgumentParser(description=__doc__)
  #parser.add_argument('host')
  #args = parser.parse_args()

  x509cert = crypto.load_certificate(crypto.FILETYPE_PEM, sys.stdin.read())

  table = [
    ['subject', stringify_subject(x509cert.get_subject().get_components())],
    ['issuer', stringify_subject(x509cert.get_issuer().get_components())],
    ['notBefore', dateutil.parser.parse(x509cert.get_notBefore())],
    ['notAfter', dateutil.parser.parse(x509cert.get_notAfter())],
    ['serial', x509cert.get_serial_number()],
    ['version', stringify_version(x509cert.get_version())],
    #['extensions', x509cert.get_extension_count()],
  ]

  for idx in range(0, x509cert.get_extension_count()):
    ext = x509cert.get_extension(idx)
    if ext.get_short_name() not in ['crlDistributionPoints', 'certificatePolicies', 'authorityInfoAccess']:
      table.append([ext.get_short_name(), str(ext)[:80]])

  print tabulate.tabulate(table, headers=['field', 'value'])

if __name__ == '__main__':
  main()



import re
import socket
import uuid

from pyasn1.codec.der.decoder import decode
from OpenSSL import crypto

from py509.asn1.san import SubjectAltName


def make_serial():
  """Make a random serial number."""
  return uuid.uuid4().int


def make_pkey(key_type=crypto.TYPE_RSA, key_bits=4096):
  """Make a public/private key pair."""
  key = crypto.PKey()
  key.generate_key(key_type, key_bits)
  return key


def make_certificate_signing_request(pkey, digest='sha512', **name):
  """Make a certificate signing request."""
  csr = crypto.X509Req()
  subj = csr.get_subject()
  subj.C = name.get('C', 'US')
  subj.ST = name.get('ST', 'CA')
  subj.L = name.get('L', 'Home')
  subj.O = name.get('O', 'Home')
  #subj.OU = name.get('OU', socket.gethostbyname(socket.getfqdn()))
  #subj.CN = name.get('CN', socket.getfqdn())
  subj.OU = name.get('OU', 'Unit')
  subj.CN = name.get('CN', 'Common')
  csr.set_pubkey(pkey)
  csr.set_version(3)
  csr.sign(pkey, digest)
  return csr


def make_certificate(csr, ca_key, ca_cert, serial, not_before, not_after, digest='sha512', version=2, exts=()):
  """Make a certificate.

  The following extensions are added to all certificates in the following order
  *before* additional extensions specified by `exts` kwarg:

    - subjectKeyIdentifier
    - authorityKeyIdentifier

  """
  crt = crypto.X509()
  crt.set_serial_number(serial)
  crt.gmtime_adj_notBefore(not_before)
  crt.gmtime_adj_notAfter(not_after)
  crt.set_issuer(ca_cert.get_subject())
  crt.set_subject(csr.get_subject())
  crt.set_pubkey(csr.get_pubkey())
  crt.set_version(version)

  crt.add_extensions([
    crypto.X509Extension(b'subjectKeyIdentifier', False, b'hash', subject=crt)])
  if ca_cert.get_subject() == crt.get_subject():
    crt.add_extensions([
      crypto.X509Extension(b'authorityKeyIdentifier', False, b'keyid:always', issuer=crt)])
  else:
    crt.add_extensions([
      crypto.X509Extension(b'authorityKeyIdentifier', False, b'keyid:always', issuer=ca_cert)])

  crt.add_extensions(exts)

  crt.sign(ca_key, digest)
  return crt


def make_certificate_authority(**name):
  """Make a certificate authority.

  A certificate authority can sign certificates. For clients to be able to
  validate certificates signed by your certificate authorithy, they must trust
  the certificate returned by this function.

  """
  key = make_pkey()
  csr = make_certificate_signing_request(key, **name)
  crt = make_certificate(csr, key, csr, make_serial(), 0, 10 * 365 * 24 * 60 * 60, exts=[crypto.X509Extension(b'basicConstraints', True, b'CA:TRUE')])
  return key, crt


def load_x509_certificates(buf):
  """Load one or multiple X.509 certificates from a buffer.

  :param buf: A buffer is an instance of `basestring` and can contain multiple
    certificates.

  """
  if not isinstance(buf, basestring):
    raise ValueError('`buf` should be an instance of `basestring` not `%s`' % type(buf))

  for pem in re.findall('(-----BEGIN CERTIFICATE-----\s(\S+\n*)+\s-----END CERTIFICATE-----\s)', buf):
    yield crypto.load_certificate(crypto.FILETYPE_PEM, pem[0])


def decode_subject_alt_name(asn1_data):
  """Decode a subject alternative name's data.

  Note, not all of the possible types are handled by this method. For a
  complete listing, see https://tools.ietf.org/html/rfc3280#section-4.2.1.7.
  Currently, only DNS names, IP addresses, and URI are supported.

  """
  for name in decode(asn1_data, asn1Spec=SubjectAltName()):
    if isinstance(name, SubjectAltName):
      for entry in range(len(name)):
        component = name.getComponentByPosition(entry)
        component_name = component.getName()
        component_data = component.getComponent().asOctets()
        if component_name == 'dNSName':
          yield bytes.decode(component_data)
        elif component_name == 'iPAddress':
          yield socket.inet_ntoa(component_data)
        elif component_name == 'uniformResourceIdentifier':
          yield bytes.decode(component_data)
        else:
          # FIXME: other types are currently not handled.
          pass



from OpenSSL import crypto

from py509.x509 import make_pkey, make_certificate_signing_request, make_certificate_authority, make_certificate, make_serial, decode_subject_alt_name


def test_make_pkey():
  key = make_pkey(key_type=crypto.TYPE_RSA, key_bits=4096)
  assert key.bits() == 4096
  assert key.type() == crypto.TYPE_RSA
  assert key.check()


def test_make_certificate_signing_request():
  pkey = make_pkey()
  csr = make_certificate_signing_request(pkey, CN='Test Name')
  assert csr.verify(pkey)
  assert csr.get_subject().CN == 'Test Name'
  assert csr.get_version() == 3


def test_make_certificate_authority():
  key, crt = make_certificate_authority(CN='Test CA')
  assert crt.get_subject().CN == 'Test CA'
  assert crt.get_extension(0).get_short_name() == b'subjectKeyIdentifier'
  assert crt.get_extension(1).get_short_name() == b'authorityKeyIdentifier'
  assert crt.get_extension(2).get_short_name() == b'basicConstraints'
  # TODO: Add an ASN.1 parser so that we can decode the data that is tucked
  # away in this extension.
  #print crt.get_extension(0).get_data()
  #print crt.get_extension(1).get_data()
  #print crt.get_extension(2).get_data()


def test_make_certificate():
  ca_key, ca_crt = make_certificate_authority(CN='Test CA')
  pkey = make_pkey()
  csr = make_certificate_signing_request(pkey, CN='Test Cert')
  crt = make_certificate(
    csr, ca_key, ca_crt,
    make_serial(), 0, 10 * 365 * 24 * 60 * 60,
    exts=[crypto.X509Extension(b'subjectAltName', True, b'IP:0.0.0.0')])
  assert crt.get_subject().CN == 'Test Cert'


def test_make_san_extensions():
  e1 = crypto.X509Extension(b'subjectAltName', True, b'IP:0.0.0.0')
  assert e1
  assert list(decode_subject_alt_name(e1.get_data())) == ['0.0.0.0']
  e2 = crypto.X509Extension(b'subjectAltName', True, b'DNS:foo.com')
  assert e2
  assert list(decode_subject_alt_name(e2.get_data())) == ['foo.com']
  e3 = crypto.X509Extension(b'subjectAltName', True, b'URI:this:is:a:uri(hello-world)')
  assert e3
  assert list(decode_subject_alt_name(e3.get_data())) == ['this:is:a:uri(hello-world)']



"""ASN.1 implementation for subjectAltNames support.

Adopted and slightly modified from the NDG HTTPS client at
https://github.com/cedadev/ndg_httpsclient, which was adopted from some legacy
pyOpenSSL extensions. Unfortunately, a first-class OpenSSL + ASN.1 library
doesn't exist, so it seems common practice to include this ASN.1 boilerplate
in your own projects.

"""

from pyasn1.type import univ, constraint, char, namedtype, tag


# The maximum allowed string length allowed.
MAX = 256


class DirectoryString(univ.Choice):
  """ASN.1 Directory string class."""
  componentType = namedtype.NamedTypes(
    namedtype.NamedType(
      'teletexString', char.TeletexString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    namedtype.NamedType(
      'printableString', char.PrintableString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    namedtype.NamedType(
      'universalString', char.UniversalString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    namedtype.NamedType(
      'utf8String', char.UTF8String().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    namedtype.NamedType(
      'bmpString', char.BMPString().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    namedtype.NamedType(
      'ia5String', char.IA5String().subtype(
        subtypeSpec=constraint.ValueSizeConstraint(1, MAX))),
    )


class AttributeValue(DirectoryString):
  """ASN.1 Attribute value."""


class AttributeType(univ.ObjectIdentifier):
  """ASN.1 Attribute type."""


class AttributeTypeAndValue(univ.Sequence):
  """ASN.1 Attribute type and value class."""
  componentType = namedtype.NamedTypes(
    namedtype.NamedType('type', AttributeType()),
    namedtype.NamedType('value', AttributeValue()),
    )


class RelativeDistinguishedName(univ.SetOf):
  """ASN.1 Realtive distinguished name."""
  componentType = AttributeTypeAndValue()


class RDNSequence(univ.SequenceOf):
  """ASN.1 RDN sequence class."""
  componentType = RelativeDistinguishedName()


class Name(univ.Choice):
  """ASN.1 name class."""
  componentType = namedtype.NamedTypes(
    namedtype.NamedType('', RDNSequence()),
    )


class Extension(univ.Sequence):
  """ASN.1 extension class."""
  componentType = namedtype.NamedTypes(
    namedtype.NamedType('extnID', univ.ObjectIdentifier()),
    namedtype.DefaultedNamedType('critical', univ.Boolean('False')),
    namedtype.NamedType('extnValue', univ.OctetString()),
    )


class Extensions(univ.SequenceOf):
  """ASN.1 extensions class."""
  componentType = Extension()
  sizeSpec = univ.SequenceOf.sizeSpec + constraint.ValueSizeConstraint(1, MAX)


class AnotherName(univ.Sequence):
  componentType = namedtype.NamedTypes(
    namedtype.NamedType('type-id', univ.ObjectIdentifier()),
    namedtype.NamedType('value', univ.Any().subtype(
      explicitTag=tag.Tag(
        tag.tagClassContext,
        tag.tagFormatSimple, 0))
    ))


class GeneralName(univ.Choice):
  """ASN.1 configuration for X.509 certificate subjectAltNames fields."""
  componentType = namedtype.NamedTypes(
    namedtype.NamedType('otherName', AnotherName().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 0))),
    namedtype.NamedType('rfc822Name', char.IA5String().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 1))),
    namedtype.NamedType('dNSName', char.IA5String().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 2))),
    # namedtype.NamedType('x400Address', ORAddress().subtype(
    #  implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 3))),
    namedtype.NamedType('directoryName', Name().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 4))),
    # namedtype.NamedType('ediPartyName', EDIPartyName().subtype(
    #   implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 5))),
    namedtype.NamedType('uniformResourceIdentifier', char.IA5String().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 6))),
    namedtype.NamedType('iPAddress', univ.OctetString().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 7))),
    namedtype.NamedType('registeredID', univ.ObjectIdentifier().subtype(
      implicitTag=tag.Tag(tag.tagClassContext, tag.tagFormatSimple, 8))),
    )


class GeneralNames(univ.SequenceOf):
  """Sequence of names for ASN.1 subjectAltNames settings."""
  componentType = GeneralName()
  sizeSpec = univ.SequenceOf.sizeSpec + constraint.ValueSizeConstraint(1, MAX)


class SubjectAltName(GeneralNames):
  """ASN.1 implementation for subjectAltNames support."""



#!/usr/bin/env python

import os
import subprocess

from setuptools import setup, find_packages, Command


entry_points = {
  'console_scripts': [
    'pyssl-diff = py509.bin.diff:main',
    'pyssl-get = py509.bin.get:main',
    'pyssl-ls = py509.bin.ls:main',
    'pyssl-verify = py509.bin.verify:main',
  ],
}


class Pex(Command):

  user_options = []

  def initialize_options(self):
    """Abstract method that is required to be overwritten"""

  def finalize_options(self):
    """Abstract method that is required to be overwritten"""

  def run(self):
    if not os.path.exists('dist/wheel-cache'):
      print('You need to create dist/wheel-cache first! You\'ll need to run the following.')
      print('  mkdir dist/wheel-cache')
      print('  pip wheel -w dist/wheel-cache')
      return
    for entry in entry_points['console_scripts']:
      name, call = tuple([_.strip() for _ in entry.split('=')])
      print('Creating {0} as {1}'.format(name, call))
      subprocess.check_call([
        'pex',
        '-r', 'py509',
        '--no-pypi',
        '--repo=dist/wheel-cache',
        '-o', name,
        '-e', call])


setup(
  name='py509',
  version='0.0.4',
  description="Framework and utility code for running public key infrastructure.",
  author='Stephen Holsapple',
  author_email='sholsapp@gmail.com',
  url='https://github.com/sholsapp/py509',
  packages=find_packages(),
  install_requires=[
    'certifi',
    'cryptography',
    'pyOpenSSL',
    'pyasn1',
    'pytest',
    'python-dateutil',
    'tabulate',
  ],
  tests_require=[
    'pytest',
  ],
  entry_points=entry_points,
  cmdclass={'pexify': Pex},
)



#!/usr/bin/env python

"""Fetch a remote host's certificate."""

import argparse
import logging
import ssl

from OpenSSL import crypto

from py509 import client


logging.basicConfig(level=logging.INFO)
log = logging.getLogger(__name__)


def main():
  parser = argparse.ArgumentParser(description=__doc__)
  parser.add_argument('host')
  args = parser.parse_args()

  x509cert = client.get_host_certificate(args.host)
  print crypto.dump_certificate(crypto.FILETYPE_PEM, x509cert)


if __name__ == '__main__':
  main()
